/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: audio_oss_out.c,v 1.3 2000/11/17 14:45:48 siggi Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#if defined(__OpenBSD__)
#include <soundcard.h>
#elif defined(__FreeBSD__)
#include <machine/soundcard.h>
#else
#include <sys/soundcard.h>
#endif
#include <sys/ioctl.h>
#include <inttypes.h>
#include "audio_out.h"
#include "../metronom.h"

/*
#define VERBOSE
*/

#define AUDIO_NUM_FRAGMENTS    15
#define AUDIO_FRAGMENT_SIZE  8192

#define DSP_TEMPLATE "/dev/dsp%d"
static char audio_dev[20];

static int audio_fd;
static uint32_t sample_rate;
static uint32_t num_channels;
static uint32_t bytes_in_buffer;
static int audio_triggered;

static uint32_t lastPTS;
static int audio_step; /* pts per 32 768 samples (sample = byte/2) */

/* 
 * pts <-> samplecount assoc 
 */ 

#define PTS_TAB_NUM 50
static uint32_t audio_pts_tab[PTS_TAB_NUM]; /* organization: ringbuffer */
static uint32_t audio_pts_pos[PTS_TAB_NUM];
static int audio_pts_num, audio_pts_start ;

/*
 * open the audio device for writing to
 */
static uint32_t ao_open(uint32_t bits, uint32_t rate, uint32_t channels)
{
  int tmp;
  int caps;
  int fsize;

  if (audio_fd > -1) {

    if (rate == sample_rate)
      return 1;

    close (audio_fd);
  }

  lastPTS = 0;
  audio_pts_num = 0;
  audio_pts_start = 0;

  audio_fd=open(audio_dev,O_WRONLY|O_NDELAY);
  if(audio_fd < 0) 
  {
    fprintf(stderr, "%s: Opening audio device %s\n",
	    strerror(errno), audio_dev);
    exit (1);
  }
  
  /* We wanted non blocking open but now put it back to normal */
  fcntl(audio_fd, F_SETFL, fcntl(audio_fd, F_GETFL)&~FNDELAY);

  tmp = channels == 2 ? 1 : 0;
  ioctl(audio_fd,SNDCTL_DSP_STEREO,&tmp);

  printf ("Audio channels : %d\n",channels);
  num_channels = channels;

  tmp = bits;
  ioctl(audio_fd,SNDCTL_DSP_SAMPLESIZE,&tmp);

  tmp = rate;
  ioctl(audio_fd,SNDCTL_DSP_SPEED, &tmp);

  sample_rate = rate;

  printf ("Audio rate : %d samples/sec\n",rate);

  if (rate != tmp) {
    printf ("\n!!!! WARNING: device only seems to support %d samples/sec on this channel\n\n",tmp);
  }

  audio_step = (uint32_t) 90000 * (uint32_t) 32768 / (rate * channels);

  printf ("audio_out : audio_step %d pts per 32768 samples\n", audio_step);

  metronom_set_audio_rate (audio_step);

  /*
   * audio buffer size handling
   */

  /*
    ioctl (audio_fd, SNDCTL_DSP_GETOSPACE, &info);

    printf ("audio buffer info : %d fragments of size %d\n",
    info.fragstotal, info.fragsize);
  */

  /* this is cheating */
  /*tmp = 256;*/

  /* tmp = 0x000F000E */; /* 15 Fragments a 2^14 = 15 * 16384 Bytes */

  tmp=0 ;
  fsize = AUDIO_FRAGMENT_SIZE;
  while (fsize>0) {
    fsize /=2;
    tmp++;
  }
  tmp--;

  tmp = (AUDIO_NUM_FRAGMENTS << 16) | tmp ;

  printf ("Audio buffer fragment info : %x\n",tmp);

  ioctl(audio_fd,SNDCTL_DSP_SETFRAGMENT,&tmp); 

  /*
   * delay output until buffer is filled, if possible
   */

  ioctl (audio_fd, SNDCTL_DSP_GETCAPS, &caps);


  if ((caps & DSP_CAP_TRIGGER) > 0) {
    
    tmp = ~PCM_ENABLE_OUTPUT;
    ioctl (audio_fd,SNDCTL_DSP_SETTRIGGER,&tmp);

    audio_triggered = 0;

  } else
    audio_triggered = 1;

  bytes_in_buffer = 0;

  return 1;
}

static void ao_put_samples(int16_t* output_samples, uint32_t num_samples, 
			   uint32_t pts_)
{

  uint32_t vpts;

#ifdef VERBOSE
  printf ("Audio : play %d samples at pts %d pos %d \n",num_samples, pts_, bytes_in_buffer);   
#endif

  write(audio_fd,output_samples, num_samples<<1);
  

  /*
   * remember vpts
   */

  vpts = metronom_got_audio_samples (pts_, num_samples);

  if (audio_pts_num >= PTS_TAB_NUM) {
      
    audio_pts_tab[audio_pts_start] = vpts;
    audio_pts_pos[audio_pts_start] = bytes_in_buffer;
    
    audio_pts_start = (audio_pts_start + 1) % PTS_TAB_NUM;

  } else {

    audio_pts_tab[audio_pts_num] = vpts;
    audio_pts_pos[audio_pts_num] = bytes_in_buffer;
    audio_pts_num++;

  }


  /*
   * step values
   */

  bytes_in_buffer += num_samples<<1;
    
  /*
   * auto-trigger audio output
   */

  if (!audio_triggered && (bytes_in_buffer > (AUDIO_NUM_FRAGMENTS * AUDIO_FRAGMENT_SIZE / 2))) {
    int tmp = PCM_ENABLE_OUTPUT;
    ioctl (audio_fd,SNDCTL_DSP_SETTRIGGER,&tmp);
    audio_triggered = 1;
  }

}

static uint32_t ao_get_current_pts (void) {

  if (!audio_triggered)
    return 0;
  else {
    int pos ;

    int i;
    uint32_t ref_pos1, ref_pos2;
    uint32_t ref_pts1, ref_pts2;
    uint32_t m,pts=0;

    count_info info;
    
    ioctl (audio_fd, SNDCTL_DSP_GETOPTR, &info);

    pos = info.bytes;
    
    /*
     * tab-based approach
     */

    if (audio_pts_num>2) {
      i = 1;

      do {
	ref_pos2 = audio_pts_pos[ (audio_pts_start+i)%PTS_TAB_NUM ];
	i++;
      } while ((i<(audio_pts_num)) && (ref_pos2<pos)) ;
      i--;

      ref_pos1 = audio_pts_pos[ (audio_pts_start+i-1)%PTS_TAB_NUM ];

#ifdef VERBOSE
      printf ("PTS_TAB: (positions) %d < %d < %d (i=%d)\n",ref_pos1,pos,ref_pos2,i);
#endif 

      if (ref_pos2<ref_pos1) {
#ifdef VERBOSE
	printf ("clock skew detected!\n");
#endif 
	i--;
	ref_pos2 = audio_pts_pos[ (audio_pts_start+i)%PTS_TAB_NUM ];
	ref_pos1 = audio_pts_pos[ (audio_pts_start+i-1)%PTS_TAB_NUM ];
#ifdef VERBOSE
	printf ("=> PTS_TAB: (positions) %d < %d < %d\n", ref_pos1,pos, ref_pos2);
#endif 
      }

      ref_pts2 = audio_pts_tab [ (audio_pts_start+i)%PTS_TAB_NUM ];
      ref_pts1 = audio_pts_tab [ (audio_pts_start+i-1)%PTS_TAB_NUM ];

#ifdef VERBOSE
	printf ("PTS_TAB: (pts)       %d < ?? < %d\n", ref_pts1,ref_pts2);
#endif 

      /*
       * linear interpolation
       */
      
      m = (ref_pts2 - ref_pts1) * 32768 / (ref_pos2 - ref_pos1);

#ifdef VERBOSE
      printf ("PTS_TAB: m=%d\n",m); 
#endif 

      pts = ref_pts1 + (pos - ref_pos1) * m / 32768;

#ifdef VERBOSE
      printf ("PTS_TAB: => pts = %d (%d sec)\n",pts,pts/90000); 
#endif 

    }
#ifdef VERBOSE
    else
      printf ("PTS_TAB: Not enough data => 0 pts\n");
#endif 

    /* 
     * check
     */
    
    if (pts < lastPTS) 
      fprintf (stderr,"Ouch! lastPTS : %d current pts : %d\n",lastPTS,pts);

    lastPTS = pts ; 


    return pts;
  }
}

static int ao_get_buf_low (void) {
  count_info info;
  int nFill;

  
  if (!audio_triggered)
    return 2;
  

  ioctl (audio_fd, SNDCTL_DSP_GETOPTR, &info);

  nFill = bytes_in_buffer - info.bytes ;

  if (nFill < (AUDIO_NUM_FRAGMENTS * AUDIO_FRAGMENT_SIZE / 8)) {
#ifdef VERBOSE
    printf ("AUDIO HELP - Buffer low %d !\n", nFill);
#endif
    return 2;
  }  else if (nFill < (AUDIO_NUM_FRAGMENTS * AUDIO_FRAGMENT_SIZE / 4))
    return 1;
  
  return 0;
}

static void ao_close(void)
{
  close(audio_fd);
  audio_triggered = 0;
  audio_fd = -1;
}


static ao_functions_t audio_ossout = {
  ao_open,
  ao_put_samples,
  ao_get_buf_low,
  ao_get_current_pts,
  ao_close,
};


ao_functions_t *audio_ossout_init (void) 
{
  int   caps;
  char  devname[] = "/dev/dsp\0\0\0";
  int   best_rate;
  int   rate ;
  int   devnum;

  /*
   * find best device driver/channel
   */

  printf ("Opening audio device...");
  devnum = 0;
  best_rate = 0;
  while (devnum<16) {
    audio_fd=open(devname,O_WRONLY|O_NDELAY);

    if (audio_fd>0) {

      /* test bitrate capability */
      
      rate = 48000;
      ioctl(audio_fd,SNDCTL_DSP_SPEED, &rate);
      if (rate>best_rate) {
	strncpy (audio_dev, devname, 19);
	best_rate = rate;
      }
      
      close (audio_fd);
    }

    sprintf(devname, DSP_TEMPLATE, devnum);
    devnum++;
  }

  /*
   * open that device
   */

  audio_fd=open(audio_dev, O_WRONLY|O_NDELAY);

  if(audio_fd < 0) 
  {
    printf("%s: Opening audio device %s\n",
	   strerror(errno), audio_dev);
    // printf("=> xine cannot live without soundcard at the moment => exit\n");
    return NULL;
  } else
    printf (" %s\n", audio_dev);

  ioctl (audio_fd, SNDCTL_DSP_GETCAPS, &caps);

  if ((caps & DSP_CAP_REALTIME) > 0) {
    printf ("audio_out   : realtime check: passed :-)\n");
  } else {
    printf ("audio_out   : realtime check: *FAILED* :-(((((\n\n");
  }

  if ((caps & DSP_CAP_TRIGGER) > 0) {
    printf ("audio_out   : trigger check : passed :-)\n");
  } else {
    printf ("audio_out   : trigger check : *FAILED* :-(((((\n");
  }

  close (audio_fd);
  audio_fd = -1;

  return &audio_ossout;
}

